@@ -40,6 +40,9 @@ def default(val, d):
 def divisible_by(numer, denom):
     return (numer % denom) == 0
 
+def safe_div(numer, denom, eps = 1e-8):
+    return numer / denom.clamp(min = eps)
+
 def cycle(dl):
     while True:
         for data in dl:
@@ -194,23 +197,19 @@ class Attention(nn.Module):
         time = None
     ):
         h = self.heads
-        has_context = exists(context)
 
-        context = default(context, x)
-
-        if x.shape[-1] != self.norm.gamma.shape[-1]:
-            print(context.shape, x.shape, self.norm.gamma.shape)
+        if exists(context):
+            context = self.norm_context(context)
 
         x = self.norm(x)
 
+        context = default(context, x)
+
         if exists(self.time_cond):
             assert exists(time)
             scale, shift = self.time_cond(time).chunk(2, dim = -1)
             x = (x * (scale + 1)) + shift
 
-        if has_context:
-            context = self.norm_context(context)
-
         qkv = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), qkv)
 
@@ -347,12 +346,14 @@ class RIN(nn.Module):
         dim_latent = None,              # will default to image dim (dim)
         num_latents = 256,              # they still had to use a fair amount of latents for good results (256), in line with the Perceiver line of papers from Deepmind
         learned_sinusoidal_dim = 16,
+        latent_token_time_cond = True,  # whether to use 1 latent token as time conditioning, or do it the adaptive layernorm way (which is highly effective as shown by some other papers "Paella" - Dominic Rampas et al.)
         **attn_kwargs
     ):
         super().__init__()
         assert divisible_by(image_size, patch_size)
         dim_latent = default(dim_latent, dim)
 
+        self.image_size = image_size
         self.channels = channels # times 2 due to self-conditioning
 
         patch_height_width = image_size // patch_size
@@ -365,11 +366,14 @@ class RIN(nn.Module):
         time_dim = dim * 4
         fourier_dim = learned_sinusoidal_dim + 1
 
+        self.latent_token_time_cond = latent_token_time_cond
+        time_output_dim = dim_latent if latent_token_time_cond else time_dim
+
         self.time_mlp = nn.Sequential(
             sinu_pos_emb,
             nn.Linear(fourier_dim, time_dim),
             nn.GELU(),
-            nn.Linear(time_dim, time_dim)
+            nn.Linear(time_dim, time_output_dim)
         )
 
         # pixels to patch and back
@@ -399,7 +403,8 @@ class RIN(nn.Module):
 
         # the main RIN body parameters  - another attention is all you need moment
 
-        attn_kwargs = {**attn_kwargs, 'time_cond_dim': time_dim}
+        if not latent_token_time_cond:
+            attn_kwargs = {**attn_kwargs, 'time_cond_dim': time_dim}
 
         self.blocks = nn.ModuleList([RINBlock(dim, dim_latent = dim_latent, latent_self_attn_depth = latent_self_attn_depth, **attn_kwargs) for _ in range(depth)])
 
@@ -430,6 +435,12 @@ class RIN(nn.Module):
         if exists(latent_self_cond):
             latents = latents + self.init_self_cond_latents(latent_self_cond)
 
+        # whether the time conditioning is to be treated as one latent token or for projecting into scale and shift for adaptive layernorm
+
+        if self.latent_token_time_cond:
+            t = rearrange(t, 'b d -> b 1 d')
+            latents = torch.cat((latents, t), dim = -2)
+
         # to patches
 
         patches = self.to_patches(x)
@@ -450,6 +461,11 @@ class RIN(nn.Module):
         if not return_latents:
             return pixels
 
+        # remove time conditioning token, if that is the settings
+
+        if self.latent_token_time_cond:
+            latents = latents[:, 1:]
+
         return pixels, latents
 
 # normalize and unnormalize image
@@ -501,8 +517,8 @@ def sigmoid_schedule(t, start = -3, end = 3, tau = 1, clamp_min = 1e-9):
 def gamma_to_alpha_sigma(gamma, scale = 1):
     return torch.sqrt(gamma) * scale, torch.sqrt(1 - gamma)
 
-def gamma_to_log_snr(gamma, eps = 1e-5):
-    return -log(gamma ** -1. - 1, eps = eps)
+def gamma_to_log_snr(gamma, scale = 1, eps = 1e-5):
+    return log(gamma * (scale ** 2) / (1 - gamma), eps = eps)
 
 # gaussian diffusion
 
@@ -512,9 +528,8 @@ class GaussianDiffusion(nn.Module):
         self,
         model: RIN,
         *,
-        image_size,
         timesteps = 1000,
-        use_ddim = False,
+        use_ddim = True,
         noise_schedule = 'sigmoid',
         objective = 'eps',
         schedule_kwargs: dict = dict(),
@@ -529,7 +544,7 @@ class GaussianDiffusion(nn.Module):
         assert objective in {'x0', 'eps'}, 'objective must be either predict x0 or noise'
         self.objective = objective
 
-        self.image_size = image_size
+        self.image_size = model.image_size
 
         if noise_schedule == "linear":
             self.gamma_schedule = simple_linear_schedule
@@ -544,7 +559,7 @@ class GaussianDiffusion(nn.Module):
 
         assert scale <= 1, 'scale must be less than or equal to 1'
         self.scale = scale
-        self.normalize_img_variance = normalize_img_variance if scale < 1 else identity
+        self.maybe_normalize_img_variance = normalize_img_variance if scale < 1 else identity
 
         # gamma schedules
 
@@ -596,8 +611,8 @@ class GaussianDiffusion(nn.Module):
 
             # get predicted x0
 
-            img = self.normalize_img_variance(img)
-            model_output, last_latents = self.model(img, noise_cond, x_start, last_latents, return_latents = True)
+            maybe_normalized_img = self.maybe_normalize_img_variance(img)
+            model_output, last_latents = self.model(maybe_normalized_img, noise_cond, x_start, last_latents, return_latents = True)
 
             # get log(snr)
 
@@ -616,7 +631,7 @@ class GaussianDiffusion(nn.Module):
                 x_start = model_output
 
             elif self.objective == 'eps':
-                x_start = (img - sigma * model_output) / alpha.clamp(min = 1e-8)
+                x_start = safe_div(img - sigma * model_output, alpha)
 
             # clip x0
 
@@ -675,8 +690,8 @@ class GaussianDiffusion(nn.Module):
 
             # predict x0
 
-            img = self.normalize_img_variance(img)
-            model_output, last_latents = self.model(img, times, x_start, last_latents, return_latents = True)
+            maybe_normalized_img = self.maybe_normalize_img_variance(img)
+            model_output, last_latents = self.model(maybe_normalized_img, times, x_start, last_latents, return_latents = True)
 
             # calculate x0 and noise
 
@@ -684,7 +699,7 @@ class GaussianDiffusion(nn.Module):
                 x_start = model_output
 
             elif self.objective == 'eps':
-                x_start = (img - sigma * model_output) / alpha.clamp(min = 1e-8)
+                x_start = safe_div(img - sigma * model_output, alpha)
 
             # clip x0
 
@@ -693,7 +708,7 @@ class GaussianDiffusion(nn.Module):
             # get predicted noise
 
             if self.objective == 'x0':
-                pred_noise = (img - alpha * x_start) / sigma.clamp(min = 1e-8)
+                pred_noise = safe_div(img - alpha * x_start, sigma)
 
             elif self.objective == 'eps':
                 pred_noise = model_output
@@ -732,7 +747,7 @@ class GaussianDiffusion(nn.Module):
 
         noised_img = alpha * img + sigma * noise
 
-        noised_img = self.normalize_img_variance(noised_img)
+        noised_img = self.maybe_normalize_img_variance(noised_img)
 
         # in the paper, they had to use a really high probability of latent self conditioning, up to 90% of the time
         # slight drawback
@@ -741,10 +756,17 @@ class GaussianDiffusion(nn.Module):
 
         if random() < self.train_prob_self_cond:
             with torch.no_grad():
-                self_cond, self_latents = self.model(noised_img, times, return_latents = True)
-                self_cond = self_cond.detach()
+                model_output, self_latents = self.model(noised_img, times, return_latents = True)
                 self_latents = self_latents.detach()
 
+                if self.objective == 'x0':
+                    self_cond = model_output
+
+                elif self.objective == 'eps':
+                    self_cond = safe_div(noised_img - sigma * model_output, alpha)
+
+                self_cond = self_cond.detach()
+
         # predict and take gradient step
 
         pred = self.model(noised_img, times, self_cond, self_latents)
@@ -870,7 +892,7 @@ class Trainer(object):
             return
 
         data = {
-            'step': self.step,
+            'step': self.step + 1,
             'model': self.accelerator.get_state_dict(self.model),
             'opt': self.opt.state_dict(),
             'ema': self.ema.state_dict(),
